--****************************************************************************
-- Software: HTTP CLASS
-- Version:  1.1
-- Date:     2005/12/31
-- Author:   Valentin Schmidt
-- Contact:  fluxus@freenet.de
-- License:  Freeware
--
-- Requirements/Dependencies: 
-- o xtra "Multiuser"
-- o POST- and PUT-UPLOADS require an xtra for reading binary files (BinFile Xtra, BinaryIO, FileIO) and the corresponding file interface script,
--   and optionaly zlibXtra xtra for speeding up the base64 encoding (otherwise lingo functions can be used)
--
-- Last Changes
-- v1.02:
-- o interface() added
-- o pNetResult (as returned by getNetResult and by responseCallback) now contains both 
--   the full HTTP status line (property "status_code") and the status code as integer (property "status_num")
-- v1.01:
-- o method DELETE added
-- 
-- To Do:
-- - maybe implement digest proxy-authentication?
-- - maybe use PRegEx Xtra for parsing results?
--
--****************************************************************************


property pMUXtra           -- Instance of Multiuser Xtra

property pServer           -- Address of server
property pPort             -- port number

property pDestination
property pData
property pMethod
property pFiles
property pMaxBytes

property pNetID -- unused
property pNetDone
property pNetResult
property pNetError
property pNetRequest

property pUserAgent
property pHttpVersion
property pAccept

property pAuth
property pProxy

-- callbacks
property pRequestCbHandler
property pRequestCbObj

property pResponseCbHandler
property pResponseCbObj

property pAcceptCookies
property pCookies
property pAdditionalHeaders

property pVerbose

-- CONST
property pCRLF

----------------------------------------
-- CONSTRUCTOR
-----------------------------------------
on new me, server, port
  if voidP(port) then port = 80
  
  pServer = server
  pPort = integer(port)
  
  -- CONST
  pCRLF = numToChar(13)&numToChar(10)
  pVERSION = "1.03"
  
  -- DEFAULTS
  pUserAgent   = "HTTP-CLASS/" & pVERSION
  pHttpVersion = "1.1"
  pAccept = "text/*, application/rtf"
  
  pVerbose = 0
  
  pNetDone = FALSE
  pAcceptCookies = FALSE
  pCookies = []
  pAdditionalHeaders = []
  
  return me
end

----------------------------------------
-- PUBLIC METHODS
----------------------------------------

----------------------------------------
-- usage: script("HTTP").interface()
----------------------------------------
on interface
  r = RETURN
  s = ""
  put "script HTTP"&r after s
  put "-- HTTP Class v1.02 (c) 2005 Valentin Schmidt"&r after s
  put "-- contact: fluxus@freenet.de"&r&r after s
  put "-- [...] = optional parameter(s)"&r&r after s
  put "-- CONSTRUCTOR:"&r after s
  put "new object me"&r&r after s
  put "-- BASIC COMMANDS:"&r after s
  put "setVerbose (me, bool)"&r after s
  put "setUserAgent (me, userAgent)"&r after s
  put "setHttpVersion(me, httpVersion)"&r after s
  put "setAuthentication (me, tUsername, tPassword)"&r after s
  put "setProxy (me, tHost, tPort[, tUsername, tPassword])"&r after s
  put "setCookieMode (me, bool)"&r&r after s
  put "-- SET CALLBACKS"&r after s
  put "setRequestCallback (me, cbHandler[, cbObj])"&r after s
  put "setResponseCallback (me, cbHandler[, cbObj])"&r&r after s
  put "-- HTTP HEADERS"&r after s
  put "addCookie (me, tName, tValue[, tPath})"&r after s
  put "ad  dHeader (me, headerStr)"&r&r after s
  put "-- SEND REQUEST"&r after s
  put "request me, tDestination[, tData, tMethod,  tFiles, tMaxBytes, tContentType]"&r&r after s
  put "-- NET LINGO INTERFACE"&r after s
  put "netDone (me)"&r after s
  put "netAbort (me)"&r after s
  put "netErrorNum (me)"&r after s
  put "netErrorString (me)"&r after s
  put "getNetRequest me"&r after s
  put "getNetResult me"
end

----------------------------------------
-- 
----------------------------------------
on setAuthentication (me, tUsername, tPassword)
  pAuth = ["username":tUsername, "password":tPassword]
end

----------------------------------------
-- tUsername and tPassword optional, if specified, used for basic authentication
----------------------------------------
on setProxy (me, tHost, tPort, tUsername, tPassword)
  pProxy = ["host":tHost, "port":integer(tPort), "username":tUsername, "password":tPassword]
end

----------------------------------------
-- 
----------------------------------------
on setRequestCallback (me, cbHandler, cbObj)
  pRequestCbHandler = symbol(cbHandler)
  pRequestCbObj = cbObj
end

----------------------------------------
-- 
----------------------------------------
on setResponseCallback (me, cbHandler, cbObj)
  pResponseCbHandler = symbol(cbHandler)
  pResponseCbObj = cbObj
end

----------------------------------------
-- 
----------------------------------------
on setCookieMode (me, bool)
  pAcceptCookies = bool
end

----------------------------------------
-- 
----------------------------------------
on setVerbose (me, bool)
  pVerbose = bool
end

----------------------------------------
-- 
----------------------------------------
on setUserAgent (me, userAgent)
  pUserAgent = userAgent
end

----------------------------------------
-- 
----------------------------------------
on setAccept (me, accept)
  pAccept = accept
end

----------------------------------------
-- at the moment: still sending the same HTTP-requests!
----------------------------------------
on setHttpVersion(me, httpVersion)
  pHttpVersion = httpVersion
end

----------------------------------------
-- 
----------------------------------------
on addCookie (me, tName, tValue, tPath)
  if voidP(tPath) then tPath="/"
  pCookies.add(["name":tName, "value":tValue, "path":tPath])  
end

----------------------------------------
--
----------------------------------------
on addHeader (me, headerKey, headerValue)
  pAdditionalHeaders.add(headerKey &": " & headerValue)
end

----------------------------------------
--
----------------------------------------
on addHeaderStr (me, headerStr)
  pAdditionalHeaders.add(headerStr)
end

----------------------------------------
-- send HTTP request
----------------------------------------
on request me, tDestination, tData, tMethod,  tFiles, tMaxBytes, tContentType
  
  if voidP(tDestination) then tDestination = "/"
  if voidP(tData) then tData = [:]
  if voidP(tMethod) then tMethod = "GET"
  if voidP(tFiles) then tFiles = [:]
  if voidP(tMaxBytes) then tMaxBytes = 16 * 1024
  
  if not(tDestination starts "/" OR tDestination starts "http://") then put "/" before tDestination
  
  pDestination = tDestination
  pData = tData
  pMethod = tMethod
  pFiles = tFiles
  pMaxBytes = tMaxBytes
  
  pNetResult = VOID
  pNetDone = FALSE
  pNetError = 0 -- ???
  
  -- initialize connection instance
  pMUXtra = xtra("multiuser").new()
  pMUXtra.setNetBufferLimits(16 * 1024 *2, tMaxBytes, 100)
  
  -- define callback handlers and error check
  errCode = pMUXtra.setNetMessageHandler(#_messageHandler, me)
  if errCode <> 0 then
    -- error defining callback handlers
    stat("Error with setNetMessageHandler")
  end if
  
  if voidP(pProxy) then
    server = pServer
    port = pPort
  else
    server = pProxy["host"]
    port = pProxy["port"]
  end if
  
  -- make connection and error check
  pNetID = uniqueID()
  errCode = pMUXtra.ConnectToNetServer(pNetID, "String", server, port, "HTTP_CLASS", 1) -- pNetID used as recipient (unused)
  
  if errCode = 0 then 
    stat( "ConnectToNetServer sent to server" )
  else
    stat( "Error sending ConnectToNetServer to server" )
  end if
  
end

----------------------------------------
-- net lingo interface
----------------------------------------
on netDone (me)
  return pNetDone
end

on netAbort (me)
  me._clearMU()
end

on netErrorNum (me)
  return pNetError -- 0 = OK!
end

on netErrorString (me)
  if ilk(pMUXtra)=#instance then 
    return pMUXtra.getNetErrorString(pNetError)
  else
    return xtra("multiuser").new().getNetErrorString(pNetError)
  end if
end

on getNetRequest me
  return pNetRequest
end

on getNetResult me
  return pNetResult
end

----------------------------------------
-- PRIVATE METHODS
----------------------------------------

----------------------------------------
-- creates http request as ["cmd":cmd, "headers":headers, "body":body]
----------------------------------------
on _netRequest (me)
  cmd = ""
  headers = []
  body = ""
  
  headers.add("Host: "&pServer&":"&pPort)
  headers.add("User-Agent:" && pUserAgent)
  --headers.add("Accept: text/*, application/rtf")
  headers.add("Accept:" && pAccept)
  headers.add("Accept-Charset: ISO-8859-1")
  repeat with h in pAdditionalHeaders
    headers.add(h)
  end repeat
  
  -- COOKIES ?
  cookie_string = ""
  repeat with c in pCookies
    if (pDestination starts c["path"]) then
      if cookie_string<>"" then put "; " after cookie_string
      put c["name"]&"="&c["value"] after cookie_string
    end if
  end repeat
  if cookie_string<>"" then headers.add("Cookie:" && cookie_string)
  
  -- AUTHENTICATION ?
  if not voidP(pAuth) then
    headers.add("Authorization: Basic" && base64encode(pAuth["username"]&":"&pAuth["password"]))
  end if
  
  -- PROXY ?
  if voidP(pProxy) then
    destination = pDestination
  else
    destination = "http://" & pServer & ":" & pPort & pDestination
    if not(voidP(pProxy["username"]) OR voidP(pProxy["password"])) then
      headers.add("Proxy-Authorization: Basic" && base64encode(pProxy["username"]&":"&pProxy["password"]))
    end if
  end if
  
  --------------------
  
  case (pMethod) of
      
    "GET", "HEAD", "TRACE", "OPTIONS", "DELETE":
      if count(pData) then put "?" & getDataString(pData) after destination
      cmd = pMethod && destination && "HTTP/"&pHttpVersion
      
    "PUT":
      cmd = "PUT" && destination && "HTTP/"&pHttpVersion
      if count(pFiles) then
        body = file_get_contents (pFiles[1], "rb")            
        headers.add("Content-Length:" && body.length)
      else
        body = ""
      end if
      
    "POST":
      cmd = "POST" && destination && "HTTP/"&pHttpVersion
      fileCnt = pFiles.count
      if fileCnt=0 then -- POST WITHOUT UPLOAD: application/x-www-form-urlencoded
        body = getDataString(pData)
        headers.add("Content-Type: application/x-www-form-urlencoded")
        headers.add("Content-Length:" && body.length)
        
      else -- POST WITH UPLOAD: multipart/form-data
        q = QUOTE
        delim = "---------------------------7547246759606"
        -- MIME BLOCK
        put "--"&delim &pCRLF after body
        put "Content-Disposition: form-data; name="&q&"MAX_FILE_SIZE"&q &pCRLF after body
        put pCRLF after body
        put pMaxBytes &pCRLF after body
        
        -- POST-VARS
        dataCnt = pData.count
        repeat with i = 1 to dataCnt
          put "--"&delim &pCRLF after body
          put "Content-Disposition: form-data; name="&q&string(pData.getPropAt(i))&q &pCRLF after body
          put pCRLF after body
          put string(pData[i]) &pCRLF after body
        end repeat
        
        od = the itemdelimiter
        the itemdelimiter = the last char of the moviepath
        repeat with i = 1 to fileCnt
          file_index = string(pFiles.getPropAt(i))
          file_path = pFiles[i]
          file_name = the last item of file_path
          file_data = file_get_contents (file_path, "rb")
          put "--"&delim &pCRLF after body
          
          -- PURE ASCII DATA:
          --put "Content-Type: text/plain; name="&q&file_name&q &pCRLF after body
          --put "Content-Transfer-Encoding: 7bit" &pCRLF after body
          --put "Content-Disposition: form-data; name="&q&file_index&q&"; filename="&q&file_name&q &pCRLF after body
          --put pCRLF after body
          --put chunk_split(file_data) &pCRLF after body
          
          -- BINARY DATA:
          put "Content-Type: application/octet-stream; name="&q&file_name&q &pCRLF after body
          put "Content-Transfer-Encoding: base64" &pCRLF after body
          put "Content-Disposition: form-data; name="&q&file_index&q&"; filename="&q&file_name&q &pCRLF after body
          put pCRLF after body
          put Base64Encode(file_data) &pCRLF after body
          
          put pCRLF after body
          put "--"&delim&"--" after body
        end repeat
        the itemdelimiter = od
        
        headers.add("Content-Type: multipart/form-data; boundary="&delim)
        headers.add("Content-Length:" && body.length)
        
      end if
      
    otherwise:
      stat( "Error: Unsupported HTTP-Command!" )
      return
      
  end case
  
  --------------------
  
  return ["cmd":cmd, "headers":headers, "body":body]
end

----------------------------------------
-- handle MU messages
----------------------------------------
on _messageHandler me
  
  -- get message from message queue
  newMsg = pMUXtra.getNetMessage()
  
  -- #recipients -> [pNetID] !
  -- put newMsg
  
  -- error check
  pNetError = newMsg.errorCode
  
  if (pNetError = 0) then
    
    -- set local variables 
    senderID = newMsg.senderID
    subject = newMsg.subject
    content = newMsg.content 
    
    case(TRUE) of:
        -- initial response from server
      (senderID = "System" and subject = "ConnectToNetServer") :
        stat("ConnectToNetServer successful")
        
        -- SEND REQUEST
        pNetRequest = me._netRequest()
        
        http_str = pNetRequest["cmd"] &pCRLF
        repeat with h in pNetRequest["headers"]
          put h &pCRLF after http_str
        end repeat
        put pCRLF after http_str
        put pNetRequest["body"] after http_str
        
        pMUXtra.SendNetMessage("system", "", http_str)
        
        -- REQUEST CALLBACK ?
        if not voidP(pRequestCbHandler) then
          case (ilk(pRequestCbObj)) of
            #sprite:  sendSprite(pRequestCbObj, pRequestCbHandler, pNetRequest)     -- no error if handler missing
            #script, #instance: call(pRequestCbHandler, pRequestCbObj, pNetRequest) -- error if handler missing
            #void: sendAllSprites(pRequestCbHandler, pNetRequest)            -- no error if handler missing
          end case
        end if
        
        dbgOut(http_str)
        
        --      (content starts "HTTP") : 
        --        dbgIn(content)
        --        me._clearMU()     
        
        
      otherwise:
        dbgIn(content)
        
        finished = 0
        if content starts "HTTP/" then
          pNetResult = me.parseResponse(content)         
          body = pNetResult["body"]          
          --if pNetResult["headers"]["Transfer-Encoding"]<>"chunked" OR body.char[body.length-6..body.length] = pCRLF&"0"&pCRLF&pCRLF then 
          if pNetResult["headers"]["Transfer-Encoding"]<>"chunked" OR body.char[body.length-4..body.length] = "0"&pCRLF&pCRLF then 
            --body = body.char[1..body.length-7] -- ???
            finished = 1
          end if
          pNetResult["body"]  = body
          
          
          -- COOKIES ?
          if pAcceptCookies then
            pos = pNetResult.headers.findPos("Set-Cookie")
            if not voidP(pos) then
              repeat while TRUE
                pCookies.add(me.parseCookieString(pNetResult.headers[pos]))
                pos=pos+1
                if pos>pNetResult.headers.count then exit repeat
                if pNetResult.headers.getPropAt(pos)<>"Set-Cookie" then exit repeat
              end repeat
            end if
          end if          
          
        else -- CHUNKS
          --if content.char[content.length-6..content.length] = pCRLF&"0"&pCRLF&pCRLF then 
          if content.char[content.length-4..content.length] = "0"&pCRLF&pCRLF then 
            content = content.char[1..content.length-7]
            finished = 1
          end if
          pNetResult["body"] = pNetResult["body"] & content
        end if
        
        
        -- FINISHED?
        -- doesn't work with footers!
        if finished then
          pNetDone = TRUE
          
          -- parse body, remove chunk length lines
          if pNetResult["headers"]["Transfer-Encoding"] = "chunked" then 
            pNetResult["body"] = me.parseRawBody(pNetResult["body"] )
          end if
          
          -- RESPONSE CALLBACK ?
          if not voidP(pResponseCbHandler) then
            case (ilk(pResponseCbObj)) of
              #sprite:  sendSprite(pResponseCbObj, pResponseCbHandler, pNetResult)     -- no error if handler missing
              #script, #instance: call(pResponseCbHandler, pResponseCbObj, pNetResult) -- error if handler missing
              #void: sendAllSprites(pResponseCbHandler, pNetResult)            -- no error if handler missing
            end case
          end if
          
          -- me._clearMU()
          
        end if
        
    end case
    
  else --> errCode <> 0
    if senderID = "System" and subject = "ConnectionProblem" then
      -- trap messages with the subject "ConnectionProblem".
      -- useful to eliminate the alert box that would otherwise appear
      stat("ConnectionProblem")
      nothing
    else
      
      errStr = pMUXtra.GetNetErrorString(pNetError)
      
      -- REQUEST CALLBACK ?
      if not voidP(pRequestCbHandler) then
        case (ilk(pRequestCbObj)) of
          #sprite:  sendSprite(pRequestCbObj, pRequestCbHandler, ["error":errStr])     -- no error if handler missing
          #script, #instance: call(pRequestCbHandler, pRequestCbObj, ["error":errStr]) -- error if handler missing
          #void: sendAllSprites(pRequestCbHandler, ["error":errStr])            -- no error if handler missing
        end case
      end if
      
      stat("ERROR:" && errStr)
      me._clearMU()
      
    end if
  end if
  
end


----------------------------------------
-- clear out netMessageHandler callbacks and void instance
----------------------------------------
on _clearMU me
  
  -- clear all callbacks by setting #handlerSymbol parameter to 0
  if objectP(pMUXtra) then
    errCode = pMUXtra.setNetMessageHandler(0, me)
    errCode = pMUXtra.setNetMessageHandler(0, me, "ConnectToNetServer")
  end if
  
  -- clear instance of the Multiuser xtra
  pMUXtra = VOID
  
  stat("CLOSE MUXtra")
end

----------------------------------------
-- UTILITIES
----------------------------------------

----------------------------------------
-- shows status information
----------------------------------------
on stat s
  if pVerbose then put "STATUS:" && s
end

----------------------------------------
-- shows complete http request sent to server
----------------------------------------
on dbgOut s
  if pVerbose then put "HTTP REQUEST:" &RETURN& s
end

----------------------------------------
--  shows complete http response received from server
----------------------------------------
on dbgIn s
  if pVerbose then put "HTTP RESPONSE:" &RETURN& s
end

----------------------------------------
-- 
----------------------------------------
on parseResponse (me, response)
  
  tmp = explode (pCRLF&pCRLF, response, 2)  
  response_headers = tmp[1]
  response_body = tmp[2]
  
  response_header_lines =  explode (pCRLF, response_headers)
  
  -- first line of headers is the HTTP response code
  http_response_line = response_header_lines[1]
  
  response_code = http_response_line
  response_code_num = integer(response_code.word[2])
  
  response_header_array = [:]
  repeat with i = 2 to response_header_lines.count
    header_line = response_header_lines[i]
    tmp = explode (": ", header_line, 2)
    header = tmp[1]
    value = tmp[2]
    --response_header_array[header] = value
    response_header_array.addProp(header, value) -- to allow e.g. multiple Set-Cookie headers!
  end repeat
  response_header_array.sort() -- ???
  
  return ["status_code":response_code, "status_num":response_code_num, "headers":response_header_array, "body":response_body]
  
end

----------------------------------------
-- parse body, remove chunk length lines (for Transfer-Encoding = "chunked")
----------------------------------------
on parseRawBody (me, raw_body)  
  body = ""
  repeat while true
    tmp = explode(pCRLF, raw_body, 2)            
    if tmp.count<2 then
      put raw_body after body
      exit repeat
    end if
    len = hex2dec( tmp[1] )
    raw_body = tmp[2]
    put raw_body.char[1..len] after body
    delete char 1 to len+2 of raw_body
  end repeat
  return body
end

----------------------------------------
-- ignores: domain, expires, secure!
----------------------------------------
on parseCookieString (me, str)
  cookie = [:]
  parts = explode("; ", str)
  tmp = explode ("=", parts[1])
  cookie["name"] = tmp[1]
  cookie["value"] = tmp[2]
  repeat with i = 2 to parts.count
    tmp = explode ("=", parts[i])
    if tmp[1]="path" then cookie["path"] = tmp[2]
  end repeat
  if voidP(cookie["path"]) then cookie["path"]="/"
  return cookie
end

